// Copyright (c) 2017 Uber Technologies, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package metrics

// Counter tracks the number of times an event has occurred
type Counter interface {
	// Inc adds the given value to the counter.
	Inc(int64)
}

// NullCounter counter that does nothing
var NullCounter Counter = nullCounter{}

type nullCounter struct{}

func (nullCounter) Inc(int64) {}
